@@ -222,7 +222,7 @@ class DSLibrary(object):
         """
         raise DSLibraryException("base class called")
 
-    def open_resource(self, resource_name: str, mode: str='rb', **kwargs) -> io.RawIOBase:
+    def open_resource(self, resource_name: str, mode: str='rb', **kwargs) -> (io.RawIOBase, bool, None):
         """
         Open a stream to or from a named file-like source/destination, or from a URI.  Each input and output should be
         given a name.  The caller selects particular data sources and associates them with these names.  The burden of
@@ -240,50 +240,68 @@ class DSLibrary(object):
             * Names starting with 'http://', 'https://', 's3://', etc., are treated as URIs of external resources.
             * All other names are expected to match an input or output defined in the metadata.
 
-        :param mode:            File read/write/append mode: 'r', 'rb', 'w', 'wb', 'a', 'ab'
+        :param mode:            File read/write/append mode: 'r', 'rb', 'w', 'wb', 'a', 'ab'.
+                                Append '?' to query whether file could be opened.
 
         :param kwargs:      Additional arguments:
                                 filesystem = opens the file through a named external filesystem engine
 
-        :returns:       A file-like object.
+        :returns:       A file-like object, or if '?' is in 'mode', either False, True or None.
         """
         uri, open_args = self._xlt_resource(resource_name, _mode="inputs" if "r" in mode else "outputs", **kwargs)
         # remove formatting arguments, etc.
         open_args.pop("format", None)
         open_args.pop("format_options", None)
         open_args.pop("dask", None)
+        test_openable = "?" in mode
         # send all outputs to mlflow if requested, but only within a run - see start_run()
         use_mlflow = self._mlflow_all and ("w" in mode or "a" in mode) and self._run_id
+        # when a 'root' folder is applied we apply it here
+        #  - any we can't allow paths that 'break out' to the parent folders
+        if "root" in kwargs:
+            if is_breakout_path(uri):
+                raise DSLibraryException(f"invalid path: {uri}")
+            uri = os.path.join(kwargs.pop("root"), uri)
         # TODO if the URI is obviously external (i.e. s3://...), or local (i.e. ./my_file), it doesn't make sense to
         #   send it to mlflow
         # if ":" in uri or uri.startswith("./"):
         #    use_mlflow = False
         if use_mlflow:
+            if test_openable:
+                return False if "a" in mode else True if "w" in mode else None
             if "a" in mode:
                 raise DSLibraryException("Append mode not available with mlflow output")
             return write_stream_with_read_on_close(w_mode=mode, r_mode='r', on_close=lambda fh: mlflow.log_artifact(fh.name, uri))
         # check for a custom protocol handler
         proto_handler = load_protocol_handler(uri, env=self._env)
         if proto_handler:
+            # TODO protocol handler may not support 'test_openable'
             return proto_handler.open_resource(uri, mode, **open_args)
         # data can be requested through a named filesystem engine/provider
         filesystem = kwargs.get("filesystem")
         if filesystem:
             open_args.pop("filesystem", None)
             fs = self.get_filesystem_connection(filesystem, for_write="r" not in mode, **open_args)
+            if test_openable:
+                if "w" in mode:
+                    return True
+                else:
+                    return fs.exists(uri)
             # TODO some of 'open_args' might be appropriate to send to open() below rather than get_filesystem_connection(), above
             return fs.open(uri, mode=mode)
-        # when a 'root' folder is applied we apply it here
-        #  - any we can't allow paths that 'break out' to the parent folders
-        if "root" in kwargs:
-            if is_breakout_path(uri):
-                raise DSLibraryException(f"invalid path: {uri}")
-            uri = os.path.join(kwargs.get("root"), uri)
         # 'uri' has the URI or path of what we need to open
         # 'mode' has the open mode
         # 'open_args' has additional arguments to pass
         return self._opener(uri, mode, **open_args)
 
+    def can_open(self, resource_name: str, mode: str = 'rb', **kwargs) -> (bool, None):
+        """
+        Test whether the given resource could be opened in the given mode.  For all parameters see open_resource().
+        """
+        if "?" not in mode:
+            mode += "?"
+        return self.open_resource(resource_name, mode, **kwargs)
+
     def open_model_binary(self, part: str=None, mode: str='rb') -> io.RawIOBase:
         """
         Read or write to the main model data store for the current model.  If there is just one piece of data, like
@@ -330,7 +348,7 @@ class DSLibrary(object):
             return connect_to_filesystem(uri=uri, for_write=for_write, **kwargs)
         if system_type == "sql":
             library = kwargs.pop("library", None)
-            return connect_to_database(uri=uri, library=library, for_write=for_write, **kwargs)
+            return connect_to_database(uri=uri, library=library, for_write=for_write, dsl=self, **kwargs)
         if system_type == "nosql":
             library = kwargs.pop("library", None)
             return connect_to_nosql(uri=uri, library=library, for_write=for_write, **kwargs)
@@ -1165,7 +1183,7 @@ class DSLibrary(object):
             if "lines" not in format_options:
                 format_options["lines"] = True
 
-    def sql_select(self, sql: str, parameters: list=None, engine: str=None, data=None, table_loader: callable=None):
+    def sql_select(self, sql: str, parameters: list=None, engine: str=None, data=None, table_loader: callable=None, **kwargs):
         """
         Run an SQL 'SELECT' statement against various sources of data.  By default, tables are just filenames, but you
         can query anything else by specifying 'engine', 'data' or 'table_loader'.
@@ -1178,6 +1196,7 @@ class DSLibrary(object):
                                     to it.  If a {str -> DataFrame} is given, the named tables will be queried.
         :param table_loader:        Allows SQL to span multiple sources.  Supply a method that accepts a tuple to
                                     specify the qualified table name, and which returns a DataFrame.
+        :param kwargs:              Additional arguments to send to dslibrary.load_dataframe().
         :return:        A DataFrame.
         """
         n_sources = bool(engine) + bool(table_loader) + bool(data is not None)
@@ -1186,10 +1205,10 @@ class DSLibrary(object):
         if not n_sources:
             # default: assume tables are filenames
             def default_table_loader(table_spec: tuple):
-                return self.load_dataframe(table_spec[-1])
+                return self.load_dataframe(table_spec[-1], **kwargs)
             table_loader = default_table_loader
         if engine:
-            return self.load_dataframe(engine, sql=sql, sql_parameters=parameters)
+            return self.load_dataframe(engine, sql=sql, sql_parameters=parameters, **kwargs)
         # a collection of tables
         # NOTE: the '.sql' package is big, so we load it here, only if it is used
         from .sql.q_pandas import sql_for_tables
@@ -1206,6 +1225,26 @@ class DSLibrary(object):
         if table_loader:
             return sql_for_tables(sql, table_loader=table_loader)
 
+    def sql_exec(self, sql: str, parameters: list=None, engine: str=None, data=None, table_loader: callable=None, **kwargs):
+        """
+        Runs a SQL write operation, i.e. CREATE TABLE, INSERT, UPDATE.
+
+        :param engine:      Name of an SQL engine, where SQL will be executed.
+
+        FIXME design this...
+          - can work against a folder using file_wr_sql/folder_db, maybe even against ABS/GCS
+          - 'table_loader' would have to be extended to support write
+          - 'data' - file_wr_sql only handles files right now, not in-memory-dataframes or {}s of dataframes
+        """
+        if engine:
+            conn = self.get_sql_connection(engine, for_write=True, **kwargs)
+            cursor = conn.cursor()
+            cursor.execute(sql, parameters)
+            cursor.close()
+            conn.close()
+            return
+        raise NotImplementedError()
+
     def load_pickled_model(self, part: str=None):
         """
         Un-pickle a saved model from the default model binary, or from one of its named parts.
