@@ -4,7 +4,7 @@ import typing
 import json
 
 
-def connect_to_database(uri: str, library: str=None, for_write: bool=False, **kwargs):
+def connect_to_database(uri: str, library: str=None, for_write: bool=False, dsl=None, **kwargs):
     """
     Try, by various means, to connect to a database.
 
@@ -12,16 +12,17 @@ def connect_to_database(uri: str, library: str=None, for_write: bool=False, **kw
     :param library:     A specific library to use, if available
     :param for_write:   Whether to enable write operations.
     :param sniff_only:  True to return a ConnectStrategy instead of a connection.
+    :param dsl:         Instance of dslibrary for access to resources, i.e. when using connect_to_folder_as_database().
     :returns:  a DBI driver
     """
     # determine flavor of connection and work out which libraries to try
     libraries = []
     if library:
         libraries.append(library)
-    if ":" not in uri:
+    if ":" not in uri or "://" in uri:
         # we can point to a folder and each file is treated as a table
         from dslibrary.utils.folder_db import connect_to_folder_as_database
-        return connect_to_folder_as_database(uri, for_write=for_write, **kwargs)
+        return connect_to_folder_as_database(uri, for_write=for_write, dsl=dsl, **kwargs)
     uri_parts = uri.split(":", maxsplit=1)
     if uri_parts[0] == "jdbc":
         uri_parts = uri_parts[1:]
